"""
this script shows implicit surface like
x*x+y*y+z*-1.0=0
Adapted from a Kevin Loney's macro for Povray,
found in his public files directory:
http://www.geocities.com/qsquared_1999/pubfiles/pubfiles.html

jmsoler, 04/2004
"""

BLENDER=1
if BLENDER==1:
    import Blender
    from Blender import NMesh

from math import *
def sqr(x):
    return x**2

def f(x,y,z):
    """
    Ro=0.45
    Ri=2.0
    P=Ri*Ri+Ro*Ro
    M=Ri*Ri-Ro*Ro
    MM=M*M
    #return x**4.0+2*(x**2)*((y**2)+z**2-P)+y**4+2*(y**2)*(z**2+M)+z**4-2*(z**2)*P+MM
    return exp(-(x**4+2*sqr(x)*(sqr(y)+sqr(z)-P)+y**4+2*sqr(y)*(sqr(z)+M)+\
           z**4-2*sqr(z)*P+MM)) +\
           exp(-(x**4+2*sqr(x)*(sqr(z)+sqr(y)-P)+z**4+2*sqr(z)*(sqr(y)+M)+\
           y**4-2*sqr(y)*P+MM)) +\
           exp(-(y**4+2*sqr(y)*(sqr(x)+sqr(z)-P)+x**4+2*sqr(x)*(sqr(z)+M)+\
           z**4-2*sqr(z)*P+MM)) - 0.7
    """
    return x*x+y*y+z*z-1.0

xmin = -2.5;
xmax = -xmin;
ymin = -2.5;
ymax = -ymin;
zmin = -2.5;
zmax = -zmin;

xsegs = 32.0;
ysegs = 32.0;
zsegs = 32.0;

xstep = (xmax - xmin) / xsegs;
ystep = (ymax - ymin) / ysegs;
zstep = (zmax - zmin) / zsegs;

vertex = 8*[[0.0,0.0,0.0]]        #Vertex coordinates of each corners of the current grid cell
value =  8*[[0.0,0.0,0.0]]        #The function values at each grid cell vertex coordinate
triCount = 0;           #Total number of triangles in the mesh

CUBE=1
LEGO=0

polycube=[]

def makeTri(a, b, c, me):
   if BLENDER==1:
        #print a
        v1=NMesh.Vert(a[0],a[1],a[2]); me.verts.append(v1)
        v1=NMesh.Vert(b[0],b[1],b[2]); me.verts.append(v1)
        v1=NMesh.Vert(c[0],c[1],c[2]); me.verts.append(v1)

        f1=NMesh.Face()

        f1.append(me.verts[-1])
        f1.append(me.verts[-2])
        f1.append(me.verts[-3])

        me.faces.append(f1)       


def interp(a, b, A, B):
    C=3*[0.0]
    for n in range(3):
        C[n]=B[n]-A[n]
        C[n]=C[n]*((-a)/(b-a))+A[n];
    return C

def polygonizeTetra(v0, v1, v2, v3):
    global triCount
    triIdx = 0;
    tri = 3*[0.0]
    norm = 3*[0.0]
    
    if(value[v0] < 0):
        triIdx = triIdx + 1; 
    if(value[v1] < 0):
        triIdx = triIdx + 2; 
    if(value[v2] < 0):
        triIdx = triIdx + 4; 
    if(value[v3] < 0):
        triIdx = triIdx + 8; 

    if  triIdx==1 or triIdx==14:
            tri[0] = interp(value[v0], value[v1], vertex[v0], vertex[v1]);
            tri[1] = interp(value[v0], value[v2], vertex[v0], vertex[v2]);
            tri[2] = interp(value[v0], value[v3], vertex[v0], vertex[v3]);
            makeTri(tri[0], tri[1], tri[2],me)
            triCount = triCount + 1;

    elif triIdx==2 or triIdx==13:    
        tri[0] = interp(value[v1], value[v0], vertex[v1], vertex[v0]);
        tri[1] = interp(value[v1], value[v3], vertex[v1], vertex[v3]);
        tri[2] = interp(value[v1], value[v2], vertex[v1], vertex[v2]);
        makeTri(tri[0], tri[1], tri[2],me)
        triCount = triCount + 1;

    elif triIdx==3 or triIdx==12:    
        tri[0] = interp(value[v0], value[v3], vertex[v0], vertex[v3]);
        tri[1] = interp(value[v0], value[v2], vertex[v0], vertex[v2]);
        tri[2] = interp(value[v1], value[v3], vertex[v1], vertex[v3]);
        makeTri(tri[0], tri[1], tri[2],me)
        tri[0] = tri[2];
        tri[2] = tri[1];
        tri[1] = interp(value[v1], value[v2], vertex[v1], vertex[v2]);
        makeTri(tri[0], tri[1], tri[2],me)
        triCount = triCount + 2;

    elif triIdx==4 or triIdx==11:    
        tri[0] = interp(value[v2], value[v0], vertex[v2], vertex[v0]);
        tri[1] = interp(value[v2], value[v1], vertex[v2], vertex[v1]);
        tri[2] = interp(value[v2], value[v3], vertex[v2], vertex[v3]);
        makeTri(tri[0], tri[1], tri[2],me)
        triCount = triCount + 1;

    elif triIdx==5 or triIdx==10:    
        tri[0] = interp(value[v0], value[v1], vertex[v0], vertex[v1]);
        tri[1] = interp(value[v2], value[v3], vertex[v2], vertex[v3]);
        tri[2] = interp(value[v0], value[v3], vertex[v0], vertex[v3]);
        makeTri(tri[0], tri[1], tri[2], me)
        tri[2] = tri[1];
        tri[1] = interp(value[v1], value[v2], vertex[v1], vertex[v2]);
        makeTri(tri[0], tri[1], tri[2],me)
        triCount = triCount + 2;

    elif triIdx==6 or triIdx==9:
        tri[0] = interp(value[v0], value[v1], vertex[v0], vertex[v1]);
        tri[1] = interp(value[v1], value[v3], vertex[v1], vertex[v3]);
        tri[2] = interp(value[v2], value[v3], vertex[v2], vertex[v3]);
        makeTri(tri[0], tri[1], tri[2],me)
        tri[1] = interp(value[v0], value[v2], vertex[v0], vertex[v2]);
        makeTri(tri[0], tri[1], tri[2],me)
        triCount = triCount + 2;

    elif triIdx==7 or triIdx==8:
        tri[0] = interp(value[v3], value[v2], vertex[v3], vertex[v2]);
        tri[1] = interp(value[v3], value[v0], vertex[v3], vertex[v0]);
        tri[2] = interp(value[v3], value[v1], vertex[v3], vertex[v1]);
        makeTri(tri[0], tri[1], tri[2],me)
        triCount = triCount + 1;

def polygonizeCell():
    polygonizeTetra(0, 2, 3, 7)
    polygonizeTetra(0, 2, 6, 7)
    polygonizeTetra(0, 4, 6, 7)
    polygonizeTetra(0, 6, 1, 2)
    polygonizeTetra(0, 6, 1, 4)
    polygonizeTetra(5, 6, 1, 4)

def makeCell(X,Y,Z):
    global polycube
    lessThan = 0; 
    greaterThan = 0;
    i = 0;
    while(i < 8):
        value[i] = f(vertex[i][0], vertex[i][1], vertex[i][2]);
        if(value[i] < 0) :
            lessThan = lessThan + 1; 
        if(value[i] > 0) :
            greaterThan = greaterThan + 1; 
        i = i + 1;

    if(lessThan & greaterThan): 
       polygonizeCell()
       if CUBE or LEGO:
           polycube.append([X,Y,Z])


def  MAIN_boucle():    
        X = xmin; #print X,xstep
        while(X < xmax):
            Y = ymin;
            while(Y < ymax):
                Z = zmin;
                while(Z < zmax):
                    vertex[0] = [X, Y, Z];
                    vertex[1] = [X+xstep, Y, Z];
                    vertex[2] = [X+xstep, Y, Z+zstep];
                    vertex[3] = [X, Y, Z+zstep];
                    vertex[4] = [X, Y+ystep, Z];
                    vertex[5] = [X+xstep, Y+ystep, Z];
                    vertex[6] = [X+xstep, Y+ystep, Z+zstep];
                    vertex[7] = [X, Y+ystep, Z+zstep];
                    makeCell(X,Y,Z)
                    Z = Z + zstep;
                Y = Y + ystep;
            X = X + xstep;

if BLENDER==1:

   me=NMesh.GetRaw()
   MAIN_boucle()
   NMesh.PutRaw(me,'test',1)

   if CUBE==1:
      cube=[[0,1,2,3],
            [0,4,5,1], 
            [4,7,6,5],
            [1,5,6,2],
            [2,3,7,6],
            [0,3,7,4]]
             
      MEcube=NMesh.GetRaw()
      for c in polycube:
            X=c[0];Y=c[1];Z=c[2]

            vertex[0] = [X, Y, Z];
            vertex[1] = [X+xstep, Y, Z];
            vertex[2] = [X+xstep, Y, Z+zstep];
            vertex[3] = [X, Y, Z+zstep];
            vertex[4] = [X, Y+ystep, Z];
            vertex[5] = [X+xstep, Y+ystep, Z];
            vertex[6] = [X+xstep, Y+ystep, Z+zstep];
            vertex[7] = [X, Y+ystep, Z+zstep];

            for v in vertex:
                  MEcube.verts.append(NMesh.Vert(v[0],v[1],v[2])) 

            for c in cube:
                 f=NMesh.Face()
                 p=len(MEcube.verts)-len(vertex)

                 for t in range(len(c)):
                     f.append(MEcube.verts[p+c[t]]) 
                 MEcube.faces.append(f)

      NMesh.PutRaw(MEcube,'testcube')
 
   if LEGO==1:
         if Blender.Object.GetSelected()[0].getType()=='Mesh':
             scene = Blender.Scene.getCurrent ()
             Ob=Blender.Object.GetSelected()[0]
             for c in polycube:
               O = Blender.Object.New (Ob.getType())  
               O.setLocation(c)
               O.shareFrom(Ob)
               #objet=Ob.getData()
               scene.link(O)
